#include <stdio.h>
#include <libplayerc++/playerc++.h>
#include <time.h>
#include <libplayerinterface/player.h>

#define NUMOBSTACALES 6 
#define STATES 1000 


/*need to do this line in c++ only*/
using namespace PlayerCc;
const double PI = 3.14;
const double EPS = 0.0001;

enum Orientation {UP=0, DOWN=1, RIGHT=2, LEFT=3};

struct pose
{
	double x, y;
	int times[4];
};

pose model[STATES];

int poseCount = 0;
double stateRadius = 0.4;
double avoidDistance = 1.5;
int avoidTurn = 60;
pose previousState;


void PrintModel()
{
	for(int i=0;i<poseCount;i++)
		printf("%d (%f, %f) %d %d %d %d \n",i,model[i].x,model[i].y,model[i].times[0],model[i].times[1],model[i].times[2],model[i].times[3]);

}

Orientation GetOrientaionByMinTimes(int* times)
{
	int min =100;
	int minIndex = 0;

	for(int i=0;i<4;i++)
	{
		if(times[i]<min)	
		{
			min = times[i];
			minIndex = i;
		}
	}

	return minIndex;	
}

double GetRandomAngle(double l, double h)
{
 	return l + (double)rand() / ((double)RAND_MAX / (h-l));
}

double GetAngle(Orientation o)
{
	if(o == UP)
	{
		return GetRandomAngle(PI/2 - EPS, PI/2 + EPS);//pi/2;
	}

	if(o == RIGHT)
	{
		return GetRandomAngle(0-EPS, 0+EPS);
	}

	if(o == DOWN)
	{
		return GetRandomAngle(-PI/2 - EPS, -PI/2 + EPS);
	}

	return GetRandomAngle(-PI-EPS, -PI+EPS);

}

Orientation GetOrientation(double a)
{
	if(a>=PI/4 && a<3*PI/4)
	{
		return UP;
	}

	if(a>=-PI/4 && a<PI/4)
	{
		return RIGHT;
	}

	if(a>-3*PI/4 && a<-PI/4)
	{
		return DOWN;
	}
	
	return LEFT;
}

double ProcedePose(Position2dProxy& pos)
{
	double x=pos.GetXPos();
	double y=pos.GetYPos();
	double a=pos.GetYaw();
	pose currentState;
	for(int i=0;i<poseCount;i++)
	{
		if((model[i].x - x)*(model[i].x - x) + (model[i].y - y)*(model[i].y - y) < stateRadius*stateRadius)
		{
			Orientation o;
		if((model[i].x - previousState.x)*(model[i].x - previousState.x) + (model[i].y - previousState.y)*(model[i].y - previousState.y) < stateRadius*stateRadius)
			{
				o = GetOrientaionByMinTimes(model[i].times);
		
				model[i].times[o]++;
	
			}
			else
			{
				return a;
			}
			previousState = model[i];
			return GetAngle(o);
		}

	}
	model[poseCount].x=x;
	model[poseCount].y=y;
	for(int i=0;i<4;i++)
	{
		model[poseCount].times[i]=0;
	}
	model[poseCount].times[RIGHT]=1;

	previousState = model[poseCount];
	poseCount++;
		
	return a;
}



void SavePose(Position2dProxy& pos)
{
	double x=pos.GetXPos();
	double y=pos.GetYPos();
	double a=pos.GetYaw();
	Orientation o = GetOrientation(a);
		
	for(int i=0;i<poseCount;i++)
	{
		if((model[i].x - x)*(model[i].x - x) + (model[i].y - y)*(model[i].y - y) < stateRadius*stateRadius)
		{
		if((model[i].x - previousState.x)*(model[i].x - previousState.x) + (model[i].y - previousState.y)*(model[i].y - previousState.y) < stateRadius*stateRadius)
			{
				model[i].times[o]++;
	
			}
			
			previousState = model[i];
			return;
		}

	}

	model[poseCount].x = x;
	model[poseCount].y = y;
	for(int i=0;i<4;i++)
	{
		model[poseCount].times[i] = 0;
	}
	model[poseCount].times[o] = 1;

	previousState = model[poseCount];
	poseCount++;
}



void MoveObs(double *speed, double *turn, RangerProxy& laser, Position2dProxy& pos)
{
	*speed = (double)(rand()%7)/(double)10-0.2;
	*turn = rand()%121-60;


	if(laser[70] < avoidDistance)
	{
		*speed = 0;
		*turn = avoidTurn;
		return;
	}
	else if(laser[120] < avoidDistance)
	{
		*speed = 0;
		*turn = (-1)*avoidTurn;
		return;
	}
	else if( (laser[85] < avoidDistance) && (laser[95] < avoidDistance))
	{
		*speed = -0.2;
		*turn = (rand()%2 ? 1 : -1)*avoidTurn;
		return;
	}
	
}


void Move(double *speed, double *turn, RangerProxy& laser, Position2dProxy& pos, BlobfinderProxy& cam)
{
	printf("%d  ",cam.GetCount());
	
	srand ( time(NULL) );
	*speed = 0.5;
	
	PrintModel();
	if( (laser[75] < avoidDistance) || ((laser[75] < avoidDistance) && (laser[105] < avoidDistance)))
	{
		*speed = -0.5;
		*turn = (rand()%2 ? 1 : -1)*avoidTurn;
		pos.SetSpeed(*speed, dtor(*turn));
		SavePose(pos);
		return;
	} 
	else if(laser[60] < avoidDistance)
	{
		*turn = avoidTurn;
		pos.SetSpeed(*speed, dtor(*turn));
		SavePose(pos);
		return;
	}
	else if(laser[120] < avoidDistance)
	{
		*turn = (-1)*avoidTurn;
		pos.SetSpeed(*speed, dtor(*turn));
		SavePose(pos);
		return;
	}
	else if(laser[45] < avoidDistance)
	{
		*turn = avoidTurn/2;
		pos.SetSpeed(*speed, dtor(*turn));
		SavePose(pos);
		return;
	}
	else if(laser[135] < avoidDistance)
	{
		*turn = (-1)*avoidTurn/2;
		pos.SetSpeed(*speed, dtor(*turn));
		SavePose(pos);
		return;
	}
	else if(laser[30] < avoidDistance)
	{
		*turn = avoidTurn/12;
		pos.SetSpeed(*speed, dtor(*turn));
		SavePose(pos);
		return;
	}
	else if(laser[150] < avoidDistance)
	{
		*turn = (-1)*avoidTurn/12;
		pos.SetSpeed(*speed, dtor(*turn));
		SavePose(pos);
		return;
	}

	 
	*turn =-pos.GetYaw()+ProcedePose(pos); 
	pos.SetSpeed(*speed, *turn);
		
}



int main(int argc, char *argv[])
{	
	PlayerClient client("localhost", 6665);
	Position2dProxy position(&client, 0);
	RangerProxy laser(&client, 0);    
	BlobfinderProxy camera(&client, 0);    
	
    	position.RequestGeom();
   	laser.RequestGeom();

	Position2dProxy *obstaclesPositions[NUMOBSTACALES];
	RangerProxy *obstaclesLasers[NUMOBSTACALES];

	for(int i=1;i<=NUMOBSTACALES;i++)
	{
		obstaclesPositions[i-1]=new Position2dProxy(&client, i);
		obstaclesLasers[i-1]=new RangerProxy(&client, i);


		obstaclesPositions[i-1]->RequestGeom();
	   	obstaclesLasers[i-1]->RequestGeom();	
	}



	previousState.x=0;
	previousState.y=0;
	double speed, turn;
	
	double obsSpeed[NUMOBSTACALES], obsTurn[NUMOBSTACALES];
	
	//enable motors
	position.SetMotorEnable(true);
	
	while(true)
	{
 		client.Read();
		if(camera.GetCount()>0)
		{
			position.SetSpeed(0,0);
	
			for(int i=1;i<=NUMOBSTACALES;i++)
			{
				obstaclesPositions[i-1]->SetSpeed(0,0);
			}

			sleep(1);
			continue;
		}	
	
		Move(&speed, &turn, laser, position, camera);		

		srand ( time(NULL) );		
		for(int i=1;i<=NUMOBSTACALES;i++)
		{
			MoveObs(&obsSpeed[i-1], &obsTurn[i-1], *obstaclesLasers[i-1], *obstaclesPositions[i-1]);			
			
	
			obstaclesPositions[i-1]->SetSpeed(obsSpeed[i-1],dtor(obsTurn[i-1]));
		}

		sleep(1);
	}

	return 0;
}

